"""
Created 18th June 2019
"""
import logging

from spine_aws_common.log.constants import LoggingConstants
from spine_aws_common.log.formatting import create_log_line


FILE_DATE_FORMAT = "%Y%m%d%H"  # pylint:disable=duplicate-code


def _get_logging_function(log_type):
    """Get the logging function to use"""
    if log_type == LoggingConstants.LFR_AUDIT:
        return logging.getLogger(LoggingConstants.SPINE_LOGGER).audit
    if log_type == LoggingConstants.LFR_NMS:
        return logging.getLogger(LoggingConstants.SPINE_LOGGER).monitor
    if log_type == LoggingConstants.LFR_OPERATIONS:
        return logging.getLogger(LoggingConstants.SPINE_LOGGER).info
    if log_type == LoggingConstants.LFR_CRASHDUMP:
        return logging.getLogger(LoggingConstants.SPINE_LOGGER).crash

    return None


def write_to_file(log_preamble, log_text, substitution_dict, log_type, error_list=None):
    """
    Append the log line to the appropriate file
    Add traceback information if required
    """
    log_line = create_log_line(log_preamble, log_text, substitution_dict)
    if error_list:
        log_line = f"{log_line} - {error_list[0:]}"

    logging_func = _get_logging_function(log_type)
    if logging_func:
        # pylint: disable=deprecated-method
        logging_func(log_line)



"""
Created 18th June 2019
"""
import re


# SPII-15583 Match both direct Werkzeug output and Flask application log
# messages
TICKET_PATTERN = re.compile("(.*)/ticket([^ ]+)(.*)")

# SPII-22963 Ensure Tornado messages are redacted, this might also contain
# new lines
GP_PROVIDER_PATTERN = re.compile(r"(.*)'GPPROVIDER_\S+'(.*)", re.DOTALL)
# Ensure nhsNumbers in fhir messages in Tornado messages are redacted, this
# might also contain new lines
# fhir.nhs.uk/Id/nhs-number|
NHS_NUMBER_PATTERN = re.compile(r"(.*fhir.nhs.uk.*)nhs-number\%7C\d+(.*)", re.DOTALL)

# flake8: noqa: E501
# pylint:disable=line-too-long
URL_PID_PATTERN = (
    "([?&])(nhs[nN]umber|identifier|forename|surname|othername|title|suffix|dateOfBirth|"
    + "timeOfBirth|address_doornumbername_Holder_Field|address_street_Holder_Field|address_town_Holder_Field|"
    + "address_locality_Holder_Field|address_county_Holder_Field|address_postcode_Holder_Field|"
    + "address_pafKey_Holder_Field)=[^& ]+"
)
# pylint:enable=line-too-long
URL_PID_REPL = r"\1\2=___MASKED___"
URL_PID_RE_COMPILED = re.compile(URL_PID_PATTERN)


def mask_ticket(val):
    """
    Mask the ticket (eg. "/demographicspineapplication/ticket?id=123... "
    would become:
    "/demographicspineapplication/ticket ___TICKET_MASKED___ "
    """
    match = TICKET_PATTERN.match(val)
    if match:
        masked_val = match.group(1) + "/ticket ___TICKET_MASKED___ "
        if match.lastindex > 2:
            return masked_val + match.group(3)
        return masked_val
    return val


def mask_gp_provider(val):
    """
    Mask the GPProvider part as it contains PID.
    """
    match = GP_PROVIDER_PATTERN.match(val)
    if match:
        masked_val = match.group(1) + "'GPPROVIDER___MASKED___'"
        if match.lastindex > 1:
            return masked_val + match.group(2)
        return masked_val
    return val


def mask_nhs_number(val):
    """
    Mask the GPProvider part as it contains PID.
    """
    match = NHS_NUMBER_PATTERN.match(val)
    if match:
        masked_val = match.group(1) + "nhs-number%7C___MASKED___"
        if match.lastindex > 1:
            return masked_val + match.group(2)
        return masked_val
    return val


def mask_sensitive_url_data(val):
    """
    Mask any true patient-identifying data from the query string of a URL,
    including NHS number.
    This is initially based on a number allocation request in DSA rest
    """
    return URL_PID_RE_COMPILED.sub(URL_PID_REPL, val)


def mask_pid(val):
    """
    Mask any PID info
    """
    masked = mask_ticket(val)
    masked = mask_gp_provider(masked)
    masked = mask_nhs_number(masked)
    masked = mask_sensitive_url_data(masked)
    return masked


def mask_url(log_row_dict):
    """
    Mask out everything after a "ticket" or nhsNumber substring in the URL
    """

    def mask(key, val):
        """
        Mask only the relevant key values
        """
        if key in ["url", "requestUrl"]:
            return mask_pid(val)
        return val

    return {key: mask(key, value) for (key, value) in log_row_dict.items()}



"""
Cloud logging module
"""
# Set imports to absolute values to avoid confusion between identical package names
from __future__ import absolute_import, print_function

import datetime
import traceback
import os

from spine_aws_common.log.spinelogging import get_log_base_config
from spine_aws_common.log.details import get_log_details, return_level
from spine_aws_common.log.constants import LoggingConstants
from spine_aws_common.log.formatting import (
    add_default_keys,
    evaluate_log_keys,
    create_log_line,
    substitute_preamble_for_monitor,
)
from spine_aws_common.log.thirdpartylogging import SEVERITY_INPUT_MAP, LoggingAdapter
from spine_aws_common.log.masking import mask_url


# pylint: disable=wrong-import-order
import logging as pythonlogging

# pylint: enable=wrong-import-order


class Logger:
    """
    Standard class for handling logging within cloud application
    Logs need to be prepared ready for splunk in the same way as
    Spine applications
    """

    _WRITEPLACEHOLDER = True  # Should a placeholder be written into operational logs

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        additional_log_config=None,
        log_base=os.path.join(os.path.dirname(__file__), "cloudlogbase.cfg"),
        process_name="ANON",
        severity_threshold="INFO",
        internal_id=None,
    ):
        self._log_base_dict = get_log_base_config(log_base=log_base)
        if additional_log_config:
            self._log_base_dict.update(
                get_log_base_config(log_base=additional_log_config)
            )

        self.process_name = process_name
        self.internal_id = internal_id
        self.severity_threshold = severity_threshold
        self.severity_threshold_value = return_level(severity_threshold)[0]
        self.date_format = "%d/%m/%Y %H:%M:%S"

        self._log_base_cache = {}

    def set_internal_id(self, internal_id):
        """Set internal ID"""
        self.internal_id = internal_id

    def set_process_name(self, process_name):
        """Set process name"""
        self.process_name = process_name

    def write_log(
        self,
        log_reference="UTI9999",
        error_list=None,
        log_row_dict=None,
        severity_threshold_override=None,
        process_name=None,
    ):
        """
        The writing of the log allows the following information to be passed:
        :param log_reference: this should resolve to a log_reference within the logBase
        :type  log_reference: str
        :param error_list: the output of a sys.exc_info() where an exception has been
        caught
        :param log_row_dict - a dictionary of substitutions to be made against the
        logText in the log_reference
        :type log_row_dict: dict
        :param severity_threshold_override: Not normally present - allows the standard
        log level to be over-ridden for this entry
        :param process_name: Not normally present - allows the standard process_name to
        be over-ridden for this entry
        The process for writing a log file entry is:
        Lookup the log reference information in the log base
        Exit out if the log level of the log is above that at which the user is set
        to log (e.g. if it is a DEBUG log and the user level is set to INFO)
        Create an audit version of the log_row_dict containing sensitive data, and
        determine if an Audit entry is required
        """
        if log_row_dict is None:
            log_row_dict = {}

        if process_name is None:
            process_name = self.process_name

        if not self._log_base_dict:
            self._print_output(process_name, log_reference, log_row_dict, error_list)
            return None

        log_details = get_log_details(
            log_reference,
            self._log_base_dict,
            self._log_base_cache,
            pythonlogging=False,
        )
        if not log_details or not log_details.check_log_severity_for_log(
            severity_threshold_override, self.severity_threshold
        ):
            return None

        # If not provided, set empty values for internalID and sessionId
        add_default_keys(log_row_dict)
        evaluate_log_keys(log_details, log_row_dict)

        time_now = datetime.datetime.now()
        log_preamble = self._create_log_preamble(
            time_now, log_details.log_level, process_name, log_reference
        )
        log_row_dict_masked = mask_url(log_row_dict)

        if log_details.audit_log_required:
            self._write_to_cloudwatch(
                log_preamble,
                log_details.log_text,
                log_row_dict_masked,
                LoggingConstants.LFR_AUDIT,
            )
        else:
            self._write_to_cloudwatch(
                log_preamble,
                log_details.log_text,
                log_row_dict_masked,
                LoggingConstants.LFR_OPERATIONS,
            )

        if log_details.monitor_log_required:
            # Swap to Log_Level=MONITOR - will help prevent SALTing requirement
            # As Splunk may get matching CRC check for Audit and Monitor Log
            self._write_to_cloudwatch(
                substitute_preamble_for_monitor(log_preamble),
                log_details.log_text,
                log_row_dict_masked,
                LoggingConstants.LFR_NMS,
            )

        if log_details.check_log_severity_for_crashdump(
            severity_threshold_override, self.severity_threshold, error_list
        ):
            stub_log_reference = LoggingConstants.LR_CRASHDUMP
            stub_log_details = get_log_details(
                stub_log_reference,
                self._log_base_dict,
                self._log_base_cache,
                pythonlogging=False,
            )
            stub_log_preamble = self._create_log_preamble(
                time_now, stub_log_details.log_level, process_name, stub_log_reference
            )

            # Write stub crashdump to spinevfmoperations, so that non-SC cleared staff
            # can see a crash occurred
            self._write_to_cloudwatch(
                stub_log_preamble,
                stub_log_details.log_text,
                {"originalLogReference": log_reference},
                LoggingConstants.LFR_OPERATIONS,
            )

            # Write actual crashdump to spinevfmcrashdump
            self._write_to_cloudwatch(
                log_preamble,
                log_details.log_text,
                log_row_dict,
                LoggingConstants.LFR_CRASHDUMP,
                error_list,
            )

        return log_details.log_text

    @staticmethod
    def _print_output(process_name, log_reference, log_row_dict, error_list):
        """
        Print out error details as no log object
        """
        print_string = process_name + ": Log Reference of " + str(log_reference)
        print_string += " raised but insufficient logging details"
        print_string += " identified to write to file."
        print(print_string)
        print("Error details " + str(error_list))
        print("Log Parameters " + str(log_row_dict))

    def _create_log_preamble(self, time_now, log_level, process_name, log_reference):
        """
        Creates the string to form the initial part of any log message
        """
        log_timestamp_string = time_now.strftime(self.date_format) + "."
        log_timestamp_string += str(int(time_now.microsecond / 1000)).rjust(3, "0")

        log_preamble = log_timestamp_string + " Log_Level=" + log_level
        log_preamble = log_preamble + " Process=" + str(process_name)
        if self.internal_id:
            log_preamble = log_preamble + " internalID=" + self.internal_id
        return log_preamble + " logReference=" + str(log_reference)

    @staticmethod
    def _write_to_cloudwatch(
        log_preamble,
        log_text,
        substitution_dict,
        log_type,
        error_list=None,
    ):
        """
        Writes the log out to the standard out for Cloudwatch logging
        """
        log_line = create_log_line(log_preamble, log_text, substitution_dict)
        if error_list is not None:
            log_line = log_line + " - " + str(error_list[0:])
        print(log_line)

        if (
            log_type == LoggingConstants.LFR_CRASHDUMP
            and error_list
            and len(error_list) >= 3
        ):
            exception, value, trace = error_list
            formatted_exception = " ".join(
                traceback.format_exception(exception, value, trace)
            )
            exception_line = create_log_line(log_preamble, formatted_exception, {})
            print(exception_line)


def configure_logging_adapter(log_object):
    """
    Configure an adapter to allow libraries that use standard Python logging to output
    to our log files
    """
    root_logger = pythonlogging.getLogger()
    root_logger.handlers = []
    root_logger.setLevel(SEVERITY_INPUT_MAP[log_object.severity_threshold])

    adapter = LoggingAdapter(log_object)
    root_logger.addHandler(adapter)
    root_logger.propagate = False



"""Mailbox class that handles all the complexity of talking to MESH API"""
import platform
import os
from typing import NamedTuple
from hashlib import sha256
import atexit
import datetime
import hmac
import tempfile
import uuid
import json
import requests

from spine_aws_common.logger import Logger
from mesh_aws_client.mesh_common import MeshCommon


class MeshMessage(NamedTuple):
    """Named tuple for holding Mesh Message info"""

    file_name: str = None
    data: any = None
    src_mailbox: str = None
    dest_mailbox: str = None
    workflow_id: str = None
    message_id: str = None
    will_compress: bool = False


class MeshMailbox:  # pylint: disable=too-many-instance-attributes
    """Mailbox class that handles all the complexity of talking to MESH API"""

    AUTH_SCHEMA_NAME = "NHSMESH"

    MESH_CA_CERT = "MESH_CA_CERT"
    MESH_CLIENT_CERT = "MESH_CLIENT_CERT"
    MESH_CLIENT_KEY = "MESH_CLIENT_KEY"
    MESH_SHARED_KEY = "MESH_SHARED_KEY"
    MESH_URL = "MESH_URL"
    MESH_VERIFY_SSL = "MESH_VERIFY_SSL"
    MAILBOX_PASSWORD = "MAILBOX_PASSWORD"
    INBOUND_BUCKET = "INBOUND_BUCKET"
    INBOUND_FOLDER = "INBOUND_FOLDER"
    ALLOWED_SENDERS = "ALLOWED_SENDERS"
    ALLOWED_RECIPIENTS = "ALLOWED_RECIPIENTS"
    ALLOWED_WORKFLOW_IDS = "ALLOWED_WORKFLOW_IDS"

    VERSION = "1.0.0"

    def __init__(self, log_object: Logger, mailbox: str, environment: str):
        self.mailbox = mailbox
        self.environment = environment
        self.temp_dir_object = None
        self.params = {}
        self.log_object = log_object
        self.client_cert_file = None
        self.client_key_file = None
        self.ca_cert_file = None
        self.maybe_verify_ssl = True
        self.dest_mailbox = None
        self.workflow_id = None

        self._setup()
        atexit.register(self.clean_up)

    def _setup(self) -> None:
        """Get mailbox config from SSM parameter store"""
        self.log_object.write_log(
            "MESHMBOX0001",
            None,
            {"mailbox": self.mailbox, "environment": self.environment},
        )

        common_params = MeshCommon.get_ssm_params(f"/{self.environment}/mesh")
        mailbox_params = MeshCommon.get_ssm_params(
            f"/{self.environment}/mesh/mailboxes/{self.mailbox}"
        )
        self.params = {**common_params, **mailbox_params}
        self.maybe_verify_ssl = (
            self.params.get(MeshMailbox.MESH_VERIFY_SSL, False) == "True"
        )
        self._write_certs_to_files()

    def clean_up(self) -> None:
        """Clear up after use"""
        if self.client_cert_file:
            filename = self.client_cert_file.name
            self.client_cert_file.close()
            os.remove(filename)
        if self.client_key_file:
            filename = self.client_key_file.name
            self.client_key_file.close()
            os.remove(filename)
        if self.ca_cert_file:
            filename = self.ca_cert_file.name
            self.ca_cert_file.close()
            os.remove(filename)

    def get_param(self, param) -> str:
        """Shortcut to get a parameter"""
        return self.params.get(param, None)

    def _write_certs_to_files(self) -> None:
        """Write the certificates to a local file"""
        self.log_object.write_log("MESHMBOX0002", None, None)

        # pylint: disable=consider-using-with
        self.temp_dir_object = tempfile.TemporaryDirectory()
        temp_dir = self.temp_dir_object.name

        # store as temporary files for the mesh client / requests library
        self.client_cert_file = tempfile.NamedTemporaryFile(dir=temp_dir, delete=False)
        client_cert = self.params[MeshMailbox.MESH_CLIENT_CERT]
        self.client_cert_file.write(client_cert.encode("utf-8"))
        self.client_cert_file.seek(0)

        self.client_key_file = tempfile.NamedTemporaryFile(dir=temp_dir, delete=False)
        client_key = self.params[MeshMailbox.MESH_CLIENT_KEY]
        self.client_key_file.write(client_key.encode("utf-8"))
        self.client_key_file.seek(0)

        self.ca_cert_file = None
        if self.maybe_verify_ssl:
            self.ca_cert_file = tempfile.NamedTemporaryFile(dir=temp_dir, delete=False)
            ca_cert = self.params[MeshMailbox.MESH_CA_CERT]
            self.ca_cert_file.write(ca_cert.encode("utf-8"))
            self.ca_cert_file.seek(0)
        # pylint: enable=consider-using-with

    def _build_mesh_authorization_header(
        self, nonce: str = None, noncecount: int = 0
    ) -> str:
        """Generate MESH Authorization header for mailbox"""
        if not nonce:
            nonce = str(uuid.uuid4())
        timestamp = datetime.datetime.now().strftime("%Y%m%d%H%M")

        # e.g. NHSMESH AMP01HC001:bd0e2bd5-218e-41d0-83a9-73fdec414803:0:202005041305
        hmac_msg = (
            f"{self.mailbox}:{nonce}:{str(noncecount)}:"
            + f"{self.params[MeshMailbox.MAILBOX_PASSWORD]}:{timestamp}"
        )

        hash_code = hmac.HMAC(
            self.params[MeshMailbox.MESH_SHARED_KEY].encode(),
            hmac_msg.encode(),
            sha256,
        ).hexdigest()
        header = (
            f"{self.AUTH_SCHEMA_NAME} {self.mailbox}:{nonce}:{str(noncecount)}:"
            + f"{timestamp}:{hash_code}"
        )
        self.log_object.write_log("MESHMBOX0003", None, {"header": header})
        return header

    def _default_headers(self):
        """
        Build standard headers including authorization
        """
        return {
            "Authorization": self._build_mesh_authorization_header(),
        }

    def _setup_session(self) -> requests.Session:
        session = requests.Session()
        session.headers = self._default_headers()
        if self.maybe_verify_ssl:
            session.verify = self.ca_cert_file.name
        else:
            session.verify = False
        session.cert = (self.client_cert_file.name, self.client_key_file.name)
        return session

    def set_destination_and_workflow(self, dest_mailbox, workflow_id) -> None:
        """Set destination mailbox and workflow_id"""
        self.dest_mailbox = dest_mailbox
        self.workflow_id = workflow_id

    def handshake(self) -> int:
        """
        Do an authenticated handshake with the MESH server
        """
        session = self._setup_session()
        session.headers[
            "Mex-ClientVersion"
        ] = f"AWS Serverless MESH Client={MeshMailbox.VERSION}"
        session.headers["Mex-OSArchitecture"] = platform.machine()
        session.headers["Mex-OSName"] = platform.system()
        session.headers["Mex-OSVersion"] = platform.release()

        mesh_url = self.params[MeshMailbox.MESH_URL]
        url = f"{mesh_url}/messageexchange/{self.mailbox}"
        response = session.get(url)
        self.log_object.write_log(
            "MESHMBOX0004", None, {"http_status": response.status_code}
        )
        return response.status_code

    def authenticate(self) -> int:
        """
        Povided for compatibility
        """
        return self.handshake()

    def send_chunk(
        self,
        mesh_message_object: MeshMessage,
        number_of_chunks: int = 1,
        chunk_num: int = 1,
    ):
        """Send a chunk from a stream"""
        # override mailbox dest_mailbox if provided in message_object
        session = self._setup_session()
        session.headers["Mex-From"] = mesh_message_object.src_mailbox
        session.headers["Mex-To"] = mesh_message_object.dest_mailbox
        session.headers["Mex-WorkflowID"] = mesh_message_object.workflow_id
        session.headers["Mex-FileName"] = mesh_message_object.file_name
        session.headers["Mex-Chunk-Range"] = f"{chunk_num}:{number_of_chunks}"
        session.headers["Content-Type"] = "application/octet-stream"
        session.headers["Mex-Content-Encrypted"] = "N"
        if mesh_message_object.will_compress:
            session.headers["Content-Encoding"] = "gzip"
            session.headers["Mex-Content-Compress"] = "Y"
            session.headers["Mex-Content-Compressed"] = "Y"

        mesh_url = self.params[MeshMailbox.MESH_URL]
        if chunk_num == 1:
            session.headers["Mex-MessageType"] = "DATA"
            url = f"{mesh_url}/messageexchange/{self.mailbox}/outbox"
        else:
            url = (
                f"{mesh_url}/messageexchange/{self.mailbox}/outbox/"
                + f"{mesh_message_object.message_id}/{chunk_num}"
            )
        response = session.post(url, data=mesh_message_object.data, stream=True)
        response.raise_for_status()
        response.raw.decode_content = True
        message_id = json.loads(response.text)["messageID"]
        self.log_object.write_log(
            "MESHSEND0007",
            None,
            {
                "file": mesh_message_object.file_name,
                "http_status": response.status_code,
                "message_id": message_id,
            },
        )
        return response

    def get_chunk(self, message_id, chunk_num=1):
        """Return a response object for a MESH chunk"""
        session = self._setup_session()
        mesh_url = self.params[MeshMailbox.MESH_URL]

        # if chunk number = 1, get first part
        if chunk_num == 1:
            url = f"{mesh_url}/messageexchange/{self.mailbox}/inbox/{message_id}"
        else:
            url = (
                f"{mesh_url}/messageexchange/{self.mailbox}/inbox/{message_id}"
                + f"/{chunk_num}"
            )
        response = session.get(url, stream=True, headers={"Accept-Encoding": "gzip"})
        response.raw.decode_content = True
        return response

    def list_messages(self):
        """Return a list of messages in the mailbox in the form:
        [
            '20220610195418651944_2202CC',
            '20220613142621549393_6430C9'
        ]
        """
        session = self._setup_session()
        mesh_url = self.params[MeshMailbox.MESH_URL]
        url = f"{mesh_url}/messageexchange/{self.mailbox}/inbox"
        response = session.get(url)
        response.raise_for_status()

        response_data = json.loads(response.text)
        message_ids = response_data["messages"]
        self.log_object.write_log(
            "MESHMBOX0005",
            None,
            {
                "mailbox": self.mailbox,
                "message_count": len(message_ids),
                "http_status": response.status_code,
            },
        )
        return response, message_ids

    def acknowledge_message(self, message_id):
        """
        Acknowledge receipt of the last message from the mailbox.
        """
        session = self._setup_session()
        mesh_url = self.params[MeshMailbox.MESH_URL]
        url = (
            f"{mesh_url}/messageexchange/{self.mailbox}/inbox/{message_id}"
            f"/status/acknowledged"
        )
        response = session.put(url)
        self.log_object.write_log(
            "MESHMBOX0006",
            None,
            {"message_id": message_id, "http_status": response.status_code},
        )
        return response



"""
Base Eventbridge Lambda application
"""
from aws_lambda_powertools.utilities.data_classes import EventBridgeEvent
from spine_aws_common.lambda_application import LambdaApplication


class EventbridgeApplication(LambdaApplication):
    """
    Base class for Eventbridge Lambda applications
    """

    EVENT_TYPE = EventBridgeEvent



"""
Base S3 Object Event Lambda application
"""
from aws_lambda_powertools.utilities.data_classes.s3_object_event import (
    S3ObjectLambdaEvent,
)
from spine_aws_common.lambda_application import LambdaApplication


class S3ObjectEventApplication(LambdaApplication):
    """
    Base class for S3 Object Event Lambda applications
    """

    EVENT_TYPE = S3ObjectLambdaEvent

    def _get_internal_id(self):
        """
        Get internalID from the event
        """
        return self._create_new_internal_id()



"""
Created 18th June 2018
"""
import configparser
import logging
import time

from spine_aws_common.log.constants import LoggingConstants

# Python Logging values commented out for reference.
AUDIT = 53
CRASH = 52
MONITOR = 51
# CRITICAL = 50
# FATAL = CRITICAL
# ERROR = 40
WARN = 31
# WARNING = 30
# WARN = WARNING
# INFO = 20
# DEBUG = 10
TRACE = 9
# NOTSET = 0

_LOG_BASE_CACHE = {}


def get_log_base_config(log_base):
    """
    Read the log base config
    - all the config is now read in at startup to move from a configparser
      object to a dictionary of tuples.
    This has a lower memory footprint (the difference in memory footprint
    is relevant to local builds - see SPII-18501)
    """
    if not log_base:
        return None

    # Use of str is required to support the case where a list of
    # log base config files have been provided
    log_base_dict = _LOG_BASE_CACHE.get(str(log_base))
    if not log_base_dict:
        log_base_config = configparser.RawConfigParser()
        log_base_config.read(log_base)
        log_base_dict = {}
        for log_ref in log_base_config.sections():
            try:
                level = log_base_config.get(log_ref, LoggingConstants.SECTION_LEVEL)
            except configparser.NoOptionError:
                level = "INFO"

            try:
                text = log_base_config.get(log_ref, LoggingConstants.SECTION_TEXT)
            except configparser.NoOptionError:
                text = log_base_config.get("UTI9999", "Log Text")

            log_base_dict[log_ref] = [level, text]

        _LOG_BASE_CACHE[str(log_base)] = log_base_dict

    return log_base_dict


class AuditFilter(logging.Filter):
    # pylint:disable=too-few-public-methods
    """
    Audit Directory Filter
    """

    def filter(self, record):
        """Only allow audit"""
        return record.levelno == AUDIT


class MonitorFilter(logging.Filter):
    # pylint:disable=too-few-public-methods
    """
    Monitor Directory Filter
    """

    def filter(self, record):
        """Only allow monitor"""
        return record.levelno == MONITOR


class CrashFilter(logging.Filter):
    # pylint:disable=too-few-public-methods
    """
    Crashdump Directory Filter
    """

    def filter(self, record):
        """Only allow crash"""
        return record.levelno == CRASH


class OperationsFilter(logging.Filter):
    # pylint:disable=too-few-public-methods
    """
    Operations Directory Filter
    """

    def filter(self, record):
        """Only allow critical and less"""
        return record.levelno <= logging.CRITICAL


class SpineLogFormatter(logging.Formatter):
    """
    Spine Log Formatter
    """

    def format_time(self, record, datefmt=None):
        """Format time"""
        converted_time = self.converter(record.created)
        time_str = time.strftime(datefmt, converted_time)
        return f"{time_str}.{int(record.msecs):0>3d}"


class SpineLogger(logging.Logger):
    """
    Spine Logger
    """

    def __init__(self, name):
        logging.Logger.__init__(self, name)

        self.log_root = None
        self.log_base = None
        self.process_name = None
        self.log_permission = None
        self.severity_threshold = None

        self.log_base_cache = {}
        self.log_base_dict = None

    def audit(self, msg, *args, **kwargs):
        """
        Log 'msg % args' with severity 'AUDIT'.
        """
        if self.isEnabledFor(AUDIT):
            self._log(AUDIT, msg, args, **kwargs)

    def monitor(self, msg, *args, **kwargs):
        """
        Log 'msg % args' with severity 'MONITOR'.
        """
        if self.isEnabledFor(MONITOR):
            self._log(MONITOR, msg, args, **kwargs)

    def crash(self, msg, *args, **kwargs):
        """
        Log 'msg % args' with severity 'CRASH'.
        """
        if self.isEnabledFor(CRASH):
            self._log(CRASH, msg, args, exc_info=True, **kwargs)

    @classmethod
    def get_process_name(cls):
        """Return the current log permission"""
        return logging.getLogger(LoggingConstants.SPINE_LOGGER).process_name

    @classmethod
    def get_log_permission(cls):
        """Return the current log permission"""
        return logging.getLogger(LoggingConstants.SPINE_LOGGER).log_permission

    @classmethod
    def get_log_base_dict(cls):
        """Return the current log base dict"""
        return logging.getLogger(LoggingConstants.SPINE_LOGGER).log_base_dict

    @classmethod
    def get_log_base_cache(cls):
        """Return the current log base cache"""
        return logging.getLogger(LoggingConstants.SPINE_LOGGER).log_base_cache

    @classmethod
    def get_severity_threshold(cls):
        """Return the current severity threshold"""
        return logging.getLogger(LoggingConstants.SPINE_LOGGER).severity_threshold


def clean_spine_logging():
    """
    Clean up the Spine Logging - setting the loggers back to how they were at
    the start. Used mainly for ensuring tests are tidied up nicely at the end.
    """
    root_logger = logging.getLogger()
    logging.root = root_logger
    logging.Logger.root = logging.root
    logging.Logger.manager = logging.Manager(logging.root)
